@@ -18,10 +18,12 @@ import array
 import binascii
 import os
 
-from os_brick.encryptors import base
-from os_brick import exception
 from oslo_concurrency import processutils
 from oslo_log import log as logging
+from oslo_log import versionutils
+
+from os_brick.encryptors import base
+from os_brick import exception
 
 LOG = logging.getLogger(__name__)
 
@@ -146,6 +148,15 @@ class CryptsetupEncryptor(base.VolumeEncryptor):
         instance is unaware of the underlying encryption due to modifying the
         original symbolic link to refer to the device mounted by dm-crypt.
         """
+        # TODO(lyarwood): Remove this encryptor and refactor the LUKS based
+        # encryptors in the U release.
+        versionutils.report_deprecated_feature(
+            LOG,
+            "The plain CryptsetupEncryptor is deprecated and will be removed "
+            "in a future release. Existing users are encouraged to retype "
+            "any existing volumes using this encryptor to the 'luks' "
+            "LuksEncryptor or 'luks2' Luks2Encryptor encryptors as soon as "
+            "possible.")
         key = self._get_key(context).get_encoded()
         passphrase = self._get_passphrase(key)
 
@@ -169,11 +180,13 @@ class CryptsetupEncryptor(base.VolumeEncryptor):
     def _close_volume(self, **kwargs):
         """Closes the device (effectively removes the dm-crypt mapping)."""
         LOG.debug("closing encrypted volume %s", self.dev_path)
-        # cryptsetup returns 4 when attempting to destroy a non-active
-        # dm-crypt device. We are going to ignore this error code to make
-        # nova deleting that instance successfully.
+        # NOTE(mdbooth): remove will return 4 (wrong device specified) if
+        # the device doesn't exist. We assume here that the caller hasn't
+        # specified the wrong device, and that it doesn't exist because it
+        # isn't open. We don't fail in this case in order to make this
+        # operation idempotent.
         self._execute('cryptsetup', 'remove', self.dev_name,
-                      run_as_root=True, check_exit_code=True,
+                      run_as_root=True, check_exit_code=[0, 4],
                       root_helper=self._root_helper)
 
     def detach_volume(self, **kwargs):
